/*
 *
 * Copyright (c) 2004 Plan B 
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * Plan B ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION. PLAN B DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * PLAN B requests users of this software to return to
 *
 *   Slick
 *   Plan B
 *   Montreal, Quebec
 *   plan_b@videotron.ca
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <limits.h>
#include <strings.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/errno.h>

int in;				/* Source File Descriptor */
int out;			/* Destination File Descriptor */
int pte;			/* Partition Table Entry Offset */
int byte_tx;			/* Bytes Transfered */
int part_dec;			/* Partition Table Entry Decimal Value */
int type_dec;			/* Partition Type Decimal Value */
const int pt_size = 66;		/* Partition Table Size in Bytes */
const int bc_size = 446;	/* Boot Code Size in Bytes */
const int sector_size = 512;	/* Sector Size in Bytes */
unsigned char buffer_in[512];	/* Source Buffer */
unsigned char buffer_out[512];	/* Destination Buffer */

int getopt_argc(int argc, char **argv);
int bc_copy(char **argv);
int pt_copy(char **argv);
int bc_print(char **argv);
int pt_print(char **argv);
int p_activate(char **argv);
int p_type(char **argv);
int ptbe_edit(char **argv);
int ptle_edit(char **argv);
int pt_sign(char **argv);
void pt_list();
void usage();

main(int argc, char **argv) {
  int c = 0;
  int flag = 1;
  if (argc == 1) { flag = 0; }
  for (c = 0; c != argc; c++) {
    if ((flag == -1) || (flag == 0)) { break; }
    if (argv[c][0] == '-') {
      flag = -1;
      if ((argv[c][1] == 'a') && (getopt_argc(argc - c, &argv[c]) == 3)) {
        p_activate(&argv[c + 1]); flag = 1; }
      if ((argv[c][1] == 'c') && (getopt_argc(argc - c, &argv[c]) == 3)) {
        bc_copy(&argv[c + 1]); flag = 1; }
      if ((argv[c][1] == 'C') && (getopt_argc(argc - c, &argv[c]) == 3)) {
        pt_copy(&argv[c + 1]); flag = 1; }
      if ((argv[c][1] == 'e') && (getopt_argc(argc - c, &argv[c]) == 9)) {
        ptbe_edit(&argv[c + 1]); flag = 1; }
      if ((argv[c][1] == 'E') && (getopt_argc(argc - c, &argv[c]) == 5)) {
        ptle_edit(&argv[c + 1]); flag = 1; }
      if ((argv[c][1] == 'h') && (getopt_argc(argc - c, &argv[c]) == 1)) {
        usage(); flag = 1; }
      if ((argv[c][1] == 'l') && (getopt_argc(argc - c, &argv[c]) == 1)) {
        pt_list(); flag = 1; }
      if ((argv[c][1] == 'p') && (getopt_argc(argc - c, &argv[c]) == 2)) {
        bc_print(&argv[c + 1]); flag = 1; }
      if ((argv[c][1] == 'P') && (getopt_argc(argc - c, &argv[c]) == 2)) {
        pt_print(&argv[c + 1]); flag = 1;}
      if ((argv[c][1] == 'S') && (getopt_argc(argc - c, &argv[c]) == 2)) {
        pt_sign(&argv[c + 1]); flag = 1; } 
      if ((argv[c][1] == 't') && (getopt_argc(argc - c, &argv[c]) == 4)) {
        p_type(&argv[c + 1]); flag = 1; }
      if  (argv[c][1] == 'u') {
        usage(); flag = 1; }
    }
  }
  if ((flag == -1) || (flag == 0)) { usage(); }
  exit (flag);
}

int getopt_argc(int argc, char **argv) {

  int c = 1;
  int opt_argc = argc;

  while (opt_argc != 1) {
    if (argv[c][0] == '-') { break; }
    else { opt_argc--; c++; }
  }
  return (c);
}

int p_activate(char **argv) {

  int c;

  part_dec = strtoul(argv[1], NULL, 10);
  if (part_dec == -1) {
    perror("\nPartition Number"); printf("\n"); exit (-1); }
  if ((part_dec < 1) || (part_dec > 4)) {
    printf("\nPartition Range: 1-4\n\n"); exit (-1); }

  in = open(argv[0], O_RDWR);
  if (in == -1) {
    perror("\nOpen Source"); printf("\n"); exit (-1); }

  byte_tx = read(in, buffer_in, sector_size);
  if (byte_tx != sector_size) {
    perror("\nRead Source Sector"); printf("\n"); exit (-1); }

  for (c = 0; c < 4; c++) { pte = bc_size + 16 * c; buffer_in[pte] = 00; }

  pte = bc_size + 16 * (part_dec - 1);
  buffer_in[pte] = 128;

  byte_tx = pwrite(in, buffer_in, sector_size, 0);
  if (byte_tx != sector_size) {
    perror("\nWrite Source Sector"); printf("\n"); exit (-1); }

  printf("\nPartition %d of %s has been activated.\n\n", part_dec, argv[0]);
}

int bc_copy(char **argv) {

  in = open(argv[0], O_RDONLY);
  if (in == -1) { perror("\nOpen Source"); printf("\n"); exit (-1); }

  byte_tx = read(in, buffer_in, sector_size);
  if (byte_tx != sector_size) {
    perror("\nRead Source Sector"); printf("\n"); exit (-1); }

  out = open(argv[1], O_RDWR|O_CREAT, S_IRUSR|S_IWUSR);
  if (out == -1) { perror("\nOpen Destination"); printf("\n"); exit (-1); }

  byte_tx = read(out, buffer_out, sector_size);
  if (byte_tx == -1) {
    perror("\nRead Destination Sector"); printf("\n"); exit (-1); }

  bcopy(buffer_in, buffer_out, bc_size);

  byte_tx = pwrite(out, buffer_out, sector_size, 0);
  if (byte_tx != sector_size) {
    perror("\nWrite Destination Sector"); printf("\n"); exit (-1); }

  printf("\nBoot Code has been transfered from %s to %s\n\n", argv[0], argv[1]);
}

int pt_copy(char **argv) {

  in = open(argv[0], O_RDONLY);
  if (in == -1) {
    perror("\nOpen Source"); printf("\n"); exit (-1); }

  byte_tx = read(in, buffer_in, sector_size);
  if (byte_tx != sector_size) {
    perror("\nRead Source Sector"); printf("\n"); exit (-1); }

  out = open(argv[1], O_RDWR|O_CREAT, S_IRUSR|S_IWUSR);
  if (out == -1) {
    perror("\nOpen Destination"); printf("\n"); exit (-1); }

  byte_tx = read(out, buffer_out, sector_size);
  if (byte_tx == -1) {
    perror("\nRead Destination Sector"); printf("\n"); exit (-1); }

  bcopy(buffer_in + bc_size, buffer_out + bc_size, pt_size);

  byte_tx = pwrite(out, buffer_out, sector_size, 0);
  if (byte_tx != sector_size) {
    perror("\nWrite Destination Sector"); printf("\n"); exit (-1); }

  printf("\nPartition Table has been transfered from %s to %s\n\n",
                                                          argv[0], argv[1]);
}

int ptbe_edit(char **argv) {

  int bc_dec;           /* Beginning Cylinder Decimal Value */
  int bh_dec;           /* Beginning Head     Decimal Value */
  int bs_dec;           /* Beginning Sector   Decimal Value */
  int ec_dec;           /* Ending    Cylinder Decimal Value */
  int eh_dec;           /* Ending    Head     Decimal Value */
  int es_dec;           /* Ending    Sector   Decimal Value */

  part_dec = strtoul(argv[1], NULL, 10);
  if (part_dec == -1) { perror("\nPartition Number"); printf("\n"); exit (-1); }
  if ((part_dec < 1) || (part_dec > 4)) {
    printf("\nPartition Range: 1-4\n\n"); exit (-1); }
  
  bc_dec = strtoul(argv[2], NULL, 10);
  if (bc_dec == -1) { perror("\nPartition Number"); printf("\n"); exit (-1); }
  ec_dec = strtoul(argv[5], NULL, 10);
  if (bh_dec == -1) { perror("\nPartition Number"); printf("\n"); exit (-1); }
  if ((bc_dec > 1024) || (ec_dec > 1024)) {
    printf("\nMaximum Cylinder Value: 1024\n\n"); exit (-1); }

  bh_dec = strtoul(argv[3], NULL, 10);
  if (bc_dec == -1) { perror("\nPartition Number"); printf("\n"); exit (-1); }
  eh_dec = strtoul(argv[6], NULL, 10);
  if (eh_dec == -1) { perror("\nPartition Number"); printf("\n"); exit (-1); }
  if ((bh_dec > 255) || (eh_dec > 255)) {
    printf("\nMaximum Head Value: 255\n\n"); exit (-1); }
 
  bs_dec = strtoul(argv[4], NULL, 10);
  if (bs_dec == -1) { perror("\nPartition Number"); printf("\n"); exit (-1); }
  es_dec = strtoul(argv[7], NULL, 10);
  if (es_dec == -1) { perror("\nPartition Number"); printf("\n"); exit (-1); }
  if ((bs_dec > 63) || (es_dec > 63)) {
    printf("\nMaximum Sector Value: 63\n\n"); exit (-1); }

  in = open(argv[0], O_RDWR);
  if (in == -1) { perror("\nOpen Source"); printf("\n"); exit (-1); }

  byte_tx = read(in, buffer_in, sector_size);
  if (byte_tx != sector_size) {
    perror("\nRead Source Sector"); printf("\n"); exit (-1); }

  pte = bc_size + 16 * (part_dec -1);
  buffer_in[pte + 1] = bh_dec;
  buffer_in[pte + 2] = bs_dec + ((bc_dec & 768) >> 2);
  buffer_in[pte + 3] = bc_dec & 255;
  buffer_in[pte + 5] = eh_dec;
  buffer_in[pte + 6] = es_dec + ((ec_dec & 768) >> 2);
  buffer_in[pte + 7] = ec_dec & 255;

  byte_tx = pwrite(in, buffer_in, sector_size, 0);
  if (byte_tx != sector_size) {
    perror("\nWrite Source Sector"); printf("\n"); exit (-1); }

  printf("\nPartition %d of %s CHS parameters have been changed to:\n\n"
       "       Beginning:\n"
       "               Cylinder: %d\n"
       "               Head:     %d\n"
       "               Sector:   %d\n"
       "       Ending:\n"
       "               Cylinder: %d\n"
       "               Head:     %d\n"
       "               Sector:   %d\n\n",
       part_dec, argv[0], bc_dec, bh_dec, bs_dec, ec_dec, eh_dec, es_dec);
}

int ptle_edit(char **argv) {

  int slba_dec;         /* Starting LBA Sector Decimal Value */
  int lbas_dec;         /* LBA Size in Sectors Decimal Value */

  part_dec = strtoul(argv[1], NULL, 10);
  if (part_dec == -1) { perror("\nPartition Number"); printf("\n"); exit (-1); }
  if ((part_dec < 1) || (part_dec > 4)) {
    printf("\nPartition Range: 1-4\n\n"); exit (-1); }

  slba_dec = strtoul(argv[2], NULL, 10);
  if (slba_dec == -1) { perror("\nStarting LBA"); printf("\n"); exit (-1); }
  if (slba_dec > 2000000000) {
    printf("\nMaximum Starting LBA Value: 2000000000\n\n"); exit (-1); }
 
  lbas_dec = strtoul(argv[3], NULL, 10);
  if (lbas_dec == -1) { perror("\nLBA Size"); printf("\n"); exit (-1); }
  if (lbas_dec > 2000000000) {
    printf("\nMaximum LBA Size value: 2000000000\n\n"); exit (-1); }

  in = open(argv[0], O_RDWR);
  if (in == -1) { perror("\nOpen Source"); printf("\n"); exit (-1); }

  byte_tx = read(in, buffer_in, sector_size);
  if (byte_tx != sector_size) {
    perror("\nRead Source Sector"); printf("\n"); exit (-1); }

  pte = bc_size + 16 * (part_dec - 1);
  buffer_in[pte +  8] =  slba_dec & 255;
  buffer_in[pte +  9] = (slba_dec & 65280) >> 8;
  buffer_in[pte + 10] = (slba_dec & 16711680) >> 16;
  buffer_in[pte + 11] =  slba_dec >> 24;
  buffer_in[pte + 12] =  lbas_dec & 255;
  buffer_in[pte + 13] = (lbas_dec & 65280) >> 8;
  buffer_in[pte + 14] = (lbas_dec & 16711680) >> 16;
  buffer_in[pte + 15] =  lbas_dec >> 24;

  byte_tx = pwrite(in, buffer_in, sector_size, 0);
  if (byte_tx != sector_size) {
    perror("\nWrite Source Sector"); printf("\n"); exit (-1); }

  printf("\nPartition %d of %s LBA parameters have been changed to:\n\n"
       "       Starting LBA: %d\n"
       "       LBA Size:     %d\n\n",
       part_dec, argv[0], slba_dec, lbas_dec);
}

int bc_print(char **argv) {

  int c;
  int bc;

  in = open(argv[0], O_RDONLY);
  if (in == -1) { perror("\nOpen Source"); printf("\n"); exit (-1); }

  byte_tx = read(in, buffer_in, sector_size);
  if (byte_tx != sector_size) {
    perror("\nRead Source Sector"); printf("\n"); exit (-1); }

  for (c = 0; c < bc_size; c++) {
    if (bc == 16) { bc = 0; printf("\n"); }
    printf("%02x", buffer_in[c]); bc++; }

  printf("\n%s Boot Code has been printed\n\n", argv[0]);
}

int pt_print(char **argv) {

  int c;

  in = open(argv[0], O_RDONLY);
  if (in == -1) { perror("\nOpen Source"); printf("\n"); exit (-1); }

  byte_tx = read(in, buffer_in, sector_size);
  if (byte_tx != sector_size) {
    perror("\nRead Source Sector"); printf("\n"); exit (-1); }

  printf("\n");
  for (c = 0; c!= 4; c++) {

    pte = bc_size + 16 * c;

    printf("Partition %2d is active %02x, type %02x,"
      " LBA start %d and LBA size %d\n",
      c + 1,
      buffer_in[pte],
      buffer_in[pte + 4],
      buffer_in[pte + 8] + (buffer_in[pte + 9] * 256)
        + (buffer_in[pte + 10] * 65536)
        + (buffer_in[pte + 11] * 16777216),
      buffer_in[pte + 12] + (buffer_in[pte + 13] * 256)
        + (buffer_in[pte + 14] * 65536)
        + (buffer_in[pte + 15] * 16777216));

    printf("        Beginning: %4d cylinder, %3d head, %2d sector\n",
      ((buffer_in[pte + 2] & 0xc0) << 2) + buffer_in[pte + 3],
        buffer_in[pte + 1], (buffer_in[pte + 2] & 0x3f));

    printf("        Ending:    %4d cylinder, %3d head, %2d sector\n\n",
      ((buffer_in[pte + 6] & 0xc0) << 2) + buffer_in[pte + 7],
        buffer_in[pte + 5], (buffer_in[pte + 6] & 0x3f));
  }
}

int p_type(char **argv) {

  part_dec = strtoul(argv[1], NULL, 10);
  if (part_dec == -1) { perror("\nPartition Number"); printf("\n"); exit (-1); }
  if ((part_dec < 1) || (part_dec > 4)) {
    printf("\nPartition Range: 1-4\n\n"); exit (-1); }

  type_dec = strtoul(argv[2], NULL, 10);
  if (type_dec == -1) { perror("\nPartition Type"); printf("\n"); exit (-1); }
  if (type_dec > 255) {
    printf("\nWrong Partition Type.\n\n"); pt_list(); exit (-1); }

  in = open(argv[0], O_RDWR);
  if (in == -1) { perror("\nOpen Source"); printf("\n"); exit (-1); }

  byte_tx = read(in, buffer_in, sector_size);
  if (byte_tx != sector_size) {
    perror("\nRead Source Sector"); printf("\n"); exit (-1); }

  pte = bc_size + 16 * (part_dec - 1);
  buffer_in[pte + 4] = type_dec;

  byte_tx = pwrite(in, buffer_in, sector_size, 0);
  if (byte_tx != sector_size) {
    perror("\nWrite Source Sector"); printf("\n"); exit (-1); }

  printf("\nPartition %d of %s type has been changed to: %d.\n\n",
    part_dec, argv[0], type_dec);
}

int pt_sign(char **argv) {

  in = open(argv[0], O_RDWR);
  if (in == -1) { perror("\nOpen Source"); printf("\n"); exit (-1); }

  byte_tx = read(in, buffer_in, sector_size);
  if (byte_tx != sector_size) {
    perror("\nRead Source Sector"); printf("\n"); exit (-1); }

  buffer_in[510] = 85;
  buffer_in[511] = 170;

  byte_tx = pwrite(in, buffer_in, sector_size, 0);
  if (byte_tx != sector_size) {
    perror("\nWrite Source Sector"); printf("\n"); exit (-1); }

  printf("\nPartition %d of %s has been signed.\n\n", part_dec, argv[0]);
}

void pt_list() {
  printf("\n"
    "00 Empty\n"
    "01 12-bit FAT primary partition\n"
    "04 16-bit FAT primary partition\n"
    "05 Extended partition\n"
    "06 BIGDOS FAT primary partition\n"
    "07 NTFS primary partition\n"
    "63 Unix System V (SCO, ISC Unix, UnixWare, Mach, GNU Hurd)\n"
    "64 Novell Netware 286, 2.xx\n"
    "65 Novell Netware 386, 3.xx or 4.xx\n"
    "69 Novell Netware 5+, Novell Netware NSS Partition\n"
    "82 Linux swap\n"
    "83 Linux native partition\n"
    "84 Hibernation partition\n"
    "85 Linux extended partition\n"
    "86 NTFS volume set\n"
    "87 NTFS volume set\n"
    "8b FAT32 volume\n"
    "8c FAT32 volume using BIOS extended INT 13h\n"
    "a0 Laptop hibernation partition\n"
    "a5 NetBSD, FreeBSD, 386BSD\n"
    "a6 OpenBSD\n"
    "a9 NetBSD\n"
    "eb BeOS\n"
    "fb VMWare File System partition\n"
    "fc VMWare Swap partition\n"
    "\n");
}

void usage() {
  printf("\n"
    "fdisk - Partition Manager\n"
    "\n"
    "SYNOPSYS\n"
    "     fdisk [-a:c:C:e:E:hlp:P:t:]"
    " [source] [destination] [partition] [type]"
    " [CHS start, CHS end] [LBA start, LBA size]\n"
    "\n"
    "        -a        Activate [partition] on [source].\n"
    "        -c        "
    "Copy Boot Code from [source] to [destination].\n"
    "        -C        "
    "Copy Partition Table from [source] to [destination].\n"
    "        -e        "
    "Edit [partition] CHS parameters on [source].\n"
    "        -E        "
    "Edit [partition] LBA parameters on [source].\n"
    "        -h        Print help.\n"
    "        -l        Print Partition Type list.\n"
    "        -p        Print Boot Code from [source].\n"
    "        -P        Print Partition Table from [source].\n"
    "        -S        Sign the Partition Table.\n"
    "        -t        Change [partition] [type] on [source].\n"
    "        -u        Print help.\n"
    "\n");
}
